/*
	FEZ: Frame Evading Zoomrects
	
	 by Douglas M. McKenna
	 	(c) Copyright 1994 All rights reserved.

	 	Mathemaesthetics, Inc.
		PO Box 298  Boulder  CO  80306-0298 USA
		
		Submitted as a hack and a half for MacHack, June 24, 1994
		Bugs are guaranteed.
	
	 	Permission is granted to use this code in any non-commercial product.
	 	Please contact the author for prior written approval for use in any
	 	commercial or shareware product.
	 	
	This file contains routines for both Standard and Frame-Evading Zoomrects.
	The frame-evading algorithms are not particularly well formed, but they
	are fun to watch.
	
	Discussion:
	
		Standard ZoomRect algorithms, such as in the Finder, simply take two
		rectangles, one the frame of a window and the other the bounds of some
		double-clickable (openable-into-the-window) item in another window, and
		interpolate a series of rectangles between them in global coordinates
		on the desktop.
		
		This is fine except for situations where (a) the item being opened or
		closed is not visible in its window, and (b) where there are intermediate
		windows covering the item.  The problem is that the metaphor of an actual
		object flying around on the desktop is broken by its starting or ending
		up outside the window in which it should be.  An object that isn't visible
		in its window, when converted to global coordinates, will be over some
		other part of the desktop or some other window, which when zoomed to, gives
		totally erroneous feedback.
		
		Even worse, the Finder will do an asynchronous zoom on the desktop while in
		the background, so that you can have a zoom whose start and end are
		totally invisible, covered by some large window in your foreground application.
		Quitting from testing an application under Metrowerks is a good example of
		this happening, and it is very annoying.  However, these routines don't
		address this problem, since it is difficult to do a proper analysis of
		windows among separate applications all on the same desktop.
		
		In order to fix some of this, the first thing is to allow more general zoom paths
		than just linear interpolations.  This is easily done using Bezier splines,
		which are easy-to-compute cubic curves that can be patched together into
		complex paths.  The other important thing needed is to keep a list of clipping
		regions that are computed based on the current windows, and maintaining a
		clipping region queue along with the standard rectangle queue that gives the
		zoom visual weight.
		
		In the case (a) above, only a single Bezier path is necessary to open or close
		a widget hidden outside the bounds of its window.  For intermediate windows,
		a simple-minded solution is used that creates amusing flying zooms.  A more
		mature solution would try to analyse the set of intermediate windows, looking
		for large holes to fly through.
		
		But then we're getting into desktop-terrain-avoidance, which algorithms have
		all been classified as military secrets :)
		
		Most of the code below was cobbled together in a very short and stressful time,
		so it could definitely be cleaned up into one set of nifty routines.
		
		Enjoy!
*/

#include "FEZ.h"
#include <math.h>

/* Prototypes for local routines */

void	InstallDesktop(void);
void	ResetDesktop(void);
short	Blender(short sc, short bc);
void	LocalToGlobalZR(ZoomRect *zoom);
void	GlobalToLocalZR(ZoomRect *zoom);
void	CenterRect(Rect *r, Rect *inside, Rect *ans);
void	Wait(short aBit);

void	MultipleSplineZoomClose(ZoomRectHandle zoomArray, ZoomRect *theWindow, ZoomRect *theWidget, short numWindows);

/* Local globals */

GrafPort deskPort;
PenState deskPen;

GrafPtr portStack[MAXPORTSTACK];		// Stack of saved GrafPtrs
short sp = 0;

Point path[MAXPATH+1];					// Where to precompute path of centers

/*
 *	Save the current port, and set the given port (if non-NIL)
 */

void PushPort(GrafPtr port)
	{
		if (sp < MAXPORTSTACK) {
			GetPort(&portStack[sp++]);
			if (port) SetPort(port);
			}
	}

/*
 *	Restore the last saved port.
 */

void PopPort()
	{
		if (sp > 0)
			SetPort(portStack[--sp]);
	}

/*
 *	Utility routines for converting between coordinates
 */

void LocalToGlobalZR(ZoomRect *zr)
	{
		Point pt;
		
		PushPort(zr->win);
		
		pt.h = zr->zoom.left; pt.v = zr->zoom.top;
		LocalToGlobal(&pt);
		zr->zoom.left = pt.h; zr->zoom.top = pt.v;
		
		pt.h = zr->zoom.right; pt.v = zr->zoom.bottom;
		LocalToGlobal(&pt);
		zr->zoom.right = pt.h; zr->zoom.bottom = pt.v;
		
		PopPort();
	}

void GlobalToLocalZR(ZoomRect *zr)
	{
		Point pt;
		
		PushPort(zr->win);
		
		pt.h = zr->zoom.left; pt.v = zr->zoom.top;
		GlobalToLocal(&pt);
		zr->zoom.left = pt.h; zr->zoom.top = pt.v;
		
		pt.h = zr->zoom.right; pt.v = zr->zoom.bottom;
		GlobalToLocal(&pt);
		zr->zoom.right = pt.h; zr->zoom.bottom = pt.v;
		
		PopPort();
	}

/*
 *	Deliver a rectangle that is the centered version of a given rectangle within
 *	another given rectangle.
 */

void CenterRect(Rect *r, Rect *inside, Rect *ans)
	{
		short rx,ry,ix,iy;

		// Use the difference between the rectangles' centers as an offset
		
		rx = (r->left + r->right) / 2;
		ry = (r->top + r->bottom) / 2;
		
		ix = (inside->right + inside->left) / 2;
		iy = (inside->bottom + inside->top) / 2;
		
		*ans = *r;
		OffsetRect(ans,ix-rx,iy-ry);
	}

/*
 *	Create a new port the size of the current desktop, and initialize
 *	its drawing environment for drawing and erasing gray zoomrects.
 */

void InstallDesktop()
	{
  		PushPort(NIL);
  		OpenPort(&deskPort);					// Sets the current port to deskPort
  		CopyRgn(LMGetGrayRgn(),deskPort.visRgn);
  		deskPort.portRect = (*LMGetGrayRgn())->rgnBBox;
  		
  		GetPenState(&deskPen);
  		
  		PenPat(&qd.gray);
  		PenMode(notPatXor);
	}

/*
 *	Restore the drawing environment in effect when InstallDesktop() was called
 */

void ResetDesktop()
	{
		SetPenState(&deskPen);
		
		ClosePort(&deskPort);
		PopPort();
	}

/*
 *	Delay by aBit ticks
 */

void Wait(short aBit)
	{
		long soon = TickCount() + aBit;
		
		while (TickCount() < soon) ;
	}

/*
 *	Draw a series of rectangles interpolated between two given rectangles.
 *	As described in TechNote 194, we draw not into WMgrPort, but into a new
 *	port covering the same area.  This code is from ancient Mac lore.
 *	Both rectangles should be specified in global screen coordinates.
 */

static Fixed fract,factor,one;

static short Blender(short smallCoord, short bigCoord)
	{
		Fixed smallFix,bigFix,tmpFix;
		
		smallFix = one * smallCoord;		/* Scale up to fixed point */
		bigFix	 = one * bigCoord;
		tmpFix	 = FixMul(fract,bigFix) + FixMul(one-fract,smallFix);
		
		return(FixRound(tmpFix));
	}

/*
 *	Zoom between two rectangles on the desktop (i.e. in global coordinates)
 *	This is the old standard zoomrect code that does linear zooms between two
 *	rectangles.  If zoomup is TRUE, the zoom goes from smallRect to bigRect;
 *	if FALSE, then it goes from bigRect back down to smallRect.  Again, this
 *	is old Mac code.
 */

void LinearZoomRect(Rect *smallRect, Rect *bigRect, int zoomUp)
	{
		short i,zoomSteps;
		Rect r1,r2,r3,r4;
		
		InstallDesktop();
		
		zoomSteps = 16;
  		one = 65536;						// fixed point constant
  		if (zoomUp) {
  			r1 = *smallRect;
  			factor = FixRatio(6,5);			// Make bigger each time
  			fract =  FixRatio(541,10000);	// 5/6 ^ 16 = 0.540877
  			}
  		 else {
  			r1 = *bigRect;
  			factor = FixRatio(5,6);			// Make smaller each time
  			fract  = one;
  			}
  		
		r2 = r1;
		r3 = r1;
		FrameRect(&r1);						// Draw initial image

		for (i=0; i<zoomSteps; i++) {
		
			r4.left   = Blender(smallRect->left,bigRect->left);
			r4.right  = Blender(smallRect->right,bigRect->right);
			r4.top	  = Blender(smallRect->top,bigRect->top);
			r4.bottom = Blender(smallRect->bottom,bigRect->bottom);
			
			FrameRect(&r4);					// Draw newest
			FrameRect(&r1);					// Erase oldest
			r1 = r2; r2 = r3; r3 = r4;		// Shift rectangle list up by 1
			
			fract = FixMul(fract,factor);	// Bump interpolation factor used by Blender
			}
		
		FrameRect(&r1);						// Erase final images
		FrameRect(&r2);
		FrameRect(&r3);
		
		ResetDesktop();
	}

/*
 *	This new routine does the same thing as LinearZoomRect(), except that
 *	instead of interpolating between the four corners of the two limiting
 *	rects, we interpolate along a line between the two limiting rects centers,
 *	and between their two sizes.  This sets us up to provide a more general
 *	routine that follows any path, instead of the implicit straight line
 *	that LinearZoomRect() would travel.  This code doesn't care about
 *	the direction of the zoom, or whether either of the rects is empty.
 */

void ZoomCenterRect(Rect *startRect, Rect *endRect, short thick)
	{
		Point startCenter,endCenter,startSize,endSize;
		Rect r1,r2,r3,r4; short zoomSteps,x,y,w,h; long i;
		PenState pen;
		
		InstallDesktop();			// Draw outside of all windows
		
  		// Get centers of the two limiting rectangles
  		
		startCenter.h = (startRect->left + startRect->right) / 2;
		startCenter.v = (startRect->top + startRect->bottom) / 2;
		endCenter.h = (endRect->left + endRect->right) / 2;
		endCenter.v = (endRect->top + endRect->bottom) / 2;
		
		// And the starting and ending half sizes
		
		startSize.h = (startRect->right - startRect->left) / 2;
		startSize.v = (startRect->bottom - startRect->top) / 2;
		endSize.h = (endRect->right - endRect->left) / 2;
		endSize.v = (endRect->bottom - endRect->top) / 2;
		
		// Fill the rectangle queue with empty rectangles so nothing gets drawn
		// initially as the queue fills up.  The queue makes for a better zoom
		// effect, since there will be more graphic weight to the zooming, and
		// the amount of time any particular rectangle is displayed is much longer.
		
  		SetRect(&r1,0,0,0,0);
		r2 = r1;
		r3 = r1;

		PenSize(thick,thick);
		
		zoomSteps = 16;
		for (i=0; i<=zoomSteps; i++) {			// loops zoomSteps+1 times
		
			// The following arithmetic is susceptable to overflow for short
			// coordinates with large magnitudes (e.g. greater than (32K/zoomSteps),
			// so we do our intermediate calculations in longs by making i a long.
			// If zoomSteps is a power of 2, it is also a lot faster to replace
			// the divide-by-zoomSteps with a right-shift by that power of 2.
			// Or you can go back and use the Fixed point stuff like the old code.
			
			// Find the i'th intermediate position along line between centers
			x = ((zoomSteps-i)*startCenter.h + i*endCenter.h) / zoomSteps;
			y = ((zoomSteps-i)*startCenter.v + i*endCenter.v) / zoomSteps;
			// And i'th intermediate sizes (interpolated linearly)
			w = ((zoomSteps-i)*startSize.h + i*endSize.h) / zoomSteps;
			h = ((zoomSteps-i)*startSize.v + i*endSize.v) / zoomSteps;
			// Build the next interpolated rectangle
			SetRect(&r4,x-w,y-h,x+w,y+h);
			// Draw the newest zooming rectangle in the queue
			FrameRect(&r4);
			// Erase (assuming xor mode) the i-3'rd previously drawn rectangle
			FrameRect(&r1);
			// Shift rectangle queue up by 1, leaving r4 ready to be redefined
			r1 = r2; r2 = r3; r3 = r4;
			
			// Use governor so processor speed doesn't affect zoom
			Wait(1);
			}
		
		// Erase last three zoomrects to empty the queue of drawn zoomrects
		
		FrameRect(&r1);
		FrameRect(&r2);
		FrameRect(&r3);
		
		ResetDesktop();			// Restore normal drawing environment
	}

/*
 *	This routine asks that the caller prepare the starting and ending ZoomRect
 *	information record, which consists of window information as well as the
 *	bounds within the window.  The coordinates of the rectangles should be global.
 *
 *	The routine decides what kind of path is needed, and uses either a straight
 *	line if possible, or it creates a curved Bezier spline path in the case that
 *	the destination rect is not visible within its window.  The path created has to
 *	look visually as if it were travelling into the window frame before changing
 *	direction and finishing the zoom to the hidden destination bounds.
 *
 *	If opening is TRUE, then start is the bounds of an item being opened within
 *	its window, and end is the bounds of the final window; otherwise, it will
 *	be the reverse.
 *
 *	This routine is only applicable when there are no intervening windows.
 *	Zooms tend to be a bit too boisterous when the widget is partially visible
 *	along the sides of large windows, however, it works pretty well in most cases.
 */

void SingleSplineZoom(ZoomRect *start, ZoomRect *end, int opening)
	{
		Point startCenter,endCenter,startSize,midSize,endSize,pt;
		Rect r1,r2,r3,r4,ans,midway,clip,content,widgetBounds;
		RgnHandle clip1,clip2,clip3,clip4,tmpClip,clipRgn;
		ZoomRect *theWidget,*theWindow;
		short zoomSteps,zoomSteps2,x,y,w,h; long i,j;
		Point2D p0,c1,c2,p3; int useCurve,turnClipOn = TRUE;
		
		clip1 = NewRgn();		// Prepare to record clipping region for rectangle queue
		clip2 = NewRgn();
		clip3 = NewRgn();
		clip4 = NewRgn();
		clipRgn = NewRgn();
		if (clipRgn == NIL) {
			if (clip4) DisposeRgn(clip4);
			if (clip3) DisposeRgn(clip3);
			if (clip2) DisposeRgn(clip2);
			if (clip1) DisposeRgn(clip1);
			return;
			}
		
		InstallDesktop();			// Draw outside of all windows
		
		GetClip(clipRgn);
		
  		// Get centers of the two limiting rectangles
  		
		startCenter.h = (start->zoom.left + start->zoom.right) / 2;
		startCenter.v = (start->zoom.top + start->zoom.bottom) / 2;
		endCenter.h = (end->zoom.left + end->zoom.right) / 2;
		endCenter.v = (end->zoom.top + end->zoom.bottom) / 2;

		// And the starting and ending half sizes
		
		startSize.h = (start->zoom.right - start->zoom.left) / 2;
		startSize.v = (start->zoom.bottom - start->zoom.top) / 2;
		endSize.h = (end->zoom.right - end->zoom.left) / 2;
		endSize.v = (end->zoom.bottom - end->zoom.top) / 2;
		
		// If either start or end rectangle is hidden in any way, then we have
		// to use a curved path to "get inside/outside" the window it's hidden in.
		// Have to convert to local coords because portRect is in local coords.
		// Since a line is a special case of a Bezier curve when the knots and
		// control points are co-linear, we can use the spline code to do a
		// linear zoom when both start and end are visible.
		
		theWidget = opening ? start : end;
		theWindow = opening ? end : start;
		
		useCurve = TRUE;
		GlobalToLocalZR(theWidget);
		if (SectRect(&theWidget->win->portRect,&theWidget->zoom,&ans))
			if (EqualRect(&ans,&theWidget->zoom)) {
				// Widget at start or end is completely visible
				// so we can use control points that force the
				// path to be a line and avoid clipping stuff
				widgetBounds = theWidget->zoom;			// Save a copy in local coords
				useCurve = FALSE;
				}
		LocalToGlobalZR(theWidget);
		
		// Get knot points on spline in global coordinates
		p0.x = startCenter.h; p0.y = startCenter.v;
		p3.x = endCenter.h; p3.y = endCenter.v;
		if (useCurve) {
		
			PushPort(theWidget->win);		// Prepare for coordinate conversions
			
			// Get content area minus the scroll bar and grow icon areas
			// Note: this assumes the window has both right and bottom scroll bars!
			content = theWidget->win->portRect;
			content.right -= SCROLLBARWIDTH;
			content.bottom -= SCROLLBARWIDTH;
			
			// Find a rectangle within the window to serve as an intermediate
			// destination for the zoom that is completely contained in the
			// visible area of the destination zoom window
			midway.left = midway.top = 0;
			midway.right = (content.right + (theWidget->zoom.right-theWidget->zoom.left)) / 2;
			midway.bottom = (content.bottom + (theWidget->zoom.bottom-theWidget->zoom.top)) / 2;
			CenterRect(&midway,&content,&midway);
			
			// If midway is larger than window, just use inset window content bounds
			if (midway.left<0 || midway.top<0 ||
					midway.right>content.right || midway.bottom>content.bottom) {
				midway = content;
				InsetRect(&midway,8,8);		// Assumes window content is at least 16 by 16
				}
			
			// Convert midway to global coords
			pt.h = midway.left; pt.v = midway.top;
			LocalToGlobal(&pt);
			midway.left = pt.h; midway.top = pt.v;
			pt.h = midway.right; pt.v = midway.bottom;
			LocalToGlobal(&pt);
			midway.right = pt.h; midway.bottom = pt.v;
			midSize.h = (midway.right - midway.left) / 2;
			midSize.v = (midway.bottom - midway.top) / 2;
			
			// Get window content as clipping rectangle in global coords
			pt.h = pt.v = 0;
			LocalToGlobal(&pt);
			clip.left = pt.h; clip.top = pt.v;
			pt.h = content.right;
			pt.v = content.bottom;
			LocalToGlobal(&pt);
			clip.right = pt.h; clip.bottom = pt.v;
			
			// Set control (tension) points to other side of window in global coordinates
			pt.h = content.right / 2;
			pt.v = content.bottom / 2;
			if (opening) {
				if (start->zoom.bottom > midway.bottom) pt.v = CONTROLINSET;
				if (start->zoom.top < midway.top) pt.v = content.bottom - CONTROLINSET;
				if (start->zoom.right > midway.right) pt.h = CONTROLINSET;
				if (start->zoom.left < midway.left) pt.h = content.right - CONTROLINSET;
				}
			 else {
				if (end->zoom.bottom > midway.bottom) pt.v = CONTROLINSET;
				if (end->zoom.top < midway.top) pt.v = content.bottom - CONTROLINSET;
				if (end->zoom.right > midway.right) pt.h = CONTROLINSET;
				if (end->zoom.left < midway.left) pt.h = content.right - CONTROLINSET;
				}
			LocalToGlobal(&pt);
			c1.x = c2.x = pt.h;
			c1.y = c2.y = pt.v;

			PopPort();
			}
		 else {
			// Set control points to 1/2 way between start and end points
			c1.x = c2.x = (p0.x + p3.x) / 2;
			c1.y = c2.y = (p0.y + p3.y) / 2;
			// midway remains undefined
			}
		
		// Precompute the spline's path, including both endpoints
		
		ComputeBezierSpline(&p0,&c1,&c2,&p3,path,MAXPATH);
				
		// Fill the rectangle queue with empty rectangles so nothing gets drawn
		// initially as the queue fills up.  The queue makes for a better zoom
		// effect, since there will be more graphic weight to the zooming, and
		// the amount of time any particular zoomrect is displayed is much longer.
		
  		SetRect(&r1,0,0,0,0);
		r2 = r1;
		r3 = r1;
		// Clipping regions are already empty
		
		PenSize(start->width,start->width);
		
		zoomSteps = MAXPATH;
		zoomSteps2 = zoomSteps/2;				// Halfway point
		
		for (i=0; i<=zoomSteps; i++) {			// loops zoomSteps+1 times
		
			x = path[i].h;
			y = path[i].v;

			if (useCurve) {
				if (i <= zoomSteps2) {
					// First half of zoom: interpolate from start to midway on desktop
					w = ((zoomSteps2-i)*startSize.h + i*midSize.h) / zoomSteps2;
					h = ((zoomSteps2-i)*startSize.v + i*midSize.v) / zoomSteps2;
					if (opening)
						ClipRect(&clip);		// Install window content clipping
					}
				 else {
				 	// Second half of zoom: interpolate from midway to end within window
				 	j = i - zoomSteps2;
					w = ((zoomSteps2-j)*midSize.h + j*endSize.h) / zoomSteps2;
					h = ((zoomSteps2-j)*midSize.v + j*endSize.v) / zoomSteps2;
					if (opening)
						SetClip(clipRgn);		// Restore original clipping region
					 else
						ClipRect(&clip);		// Install window content clipping
					}
				}
			 else {
				// Get i'th intermediate size (interpolated linearly) for whole zoom
				w = ((zoomSteps-i)*startSize.h + i*endSize.h) / zoomSteps;
				h = ((zoomSteps-i)*startSize.v + i*endSize.v) / zoomSteps;
				// No clipping needed
				}
			
			// Build and draw the next interpolated rectangle in queue
			SetRect(&r4,x-w,y-h,x+w,y+h);
			FrameRect(&r4);
			
			// Erase (assuming xor mode) the i-3'rd previously drawn rectangle
			if (useCurve) {
				GetClip(clip4);
				SetClip(clip1);
				}
			FrameRect(&r1);
			
			// Shift rectangle queue up by 1, leaving r4 ready to be redefined
			r1 = r2; r2 = r3; r3 = r4;
			if (useCurve) {
				tmpClip = clip1; clip1 = clip2; clip2 = clip3; clip3 = clip4; clip4 = tmpClip;
				}
			
			// Use governor so processor speed doesn't affect zoom
			Wait(1);
			}
		
		// Erase last three zoomrects to empty the queue of drawn zoomrects
		
		FrameRect(&r1);
		FrameRect(&r2);
		FrameRect(&r3);
		
		ResetDesktop();						// Restore normal drawing environment
		
		// Clean up our temporary storage
		DisposeRgn(clipRgn);
		DisposeRgn(clip4); DisposeRgn(clip3);
		DisposeRgn(clip2); DisposeRgn(clip1);
	}

/*
 *	This function precomputes an array of ZoomRects for doing a multiwindow frame-evading
 *	zoom on the desktop between start and end.  The delivered array will have numWindows+2
 *	entries, with the first and last entries the same as start and end.  When opening is
 *	FALSE, this function expects start->win to be visible, so that we don't have to worry
 *	about changes to the window order screwing us up when the caller eventually calls
 *	HideWindow() prior to doing the actual zoom using the array.
 *	Delivers NIL if not enough memory, and currently if opening is TRUE.
 *
 *	The ZoomArray is allocated as a relocatable block on the heap.
 */

ZoomRectHandle GetMultiZoom(ZoomRect *start, ZoomRect *end, int opening)
	{
		RgnHandle clipRgn,tmpRgn; ZoomRectHandle zoomArray=NIL;
		WindowPeek wp; int okay = FALSE; ZoomRect *theWindow,*theWidget;
		short numWindows,w,h,k,x,y; Rect ans,bounds;
		
		// It is possible to have to evade window frames when opening an item from
		// a window, because even though items being opened are usually in the frontmost
		// window, these days there can be any number of floating windows in front of
		// the frontmost document window that may need evading.  However, since this
		// testbed doesn't support floating palettes or balloon help for Fez's, I'm
		// not going to worry about it for now.  The situation is very symmetric with
		// closing, like SingleSplineZoom, but there wasn't enough time before the
		// hack contest to think about this.
		
		if (opening) return(NIL);

		theWindow = opening ? end : start;
		theWidget = opening ? start : end;
		
		clipRgn = NewRgn();
		if (clipRgn == NIL) goto cleanup;
		
		// Subtract out all visible windows that were in front of the window being closed.
		// These are typically floating palettes (or text services or balloons).
		
		CopyRgn(LMGetGrayRgn(),clipRgn);			// Start with entire visible desktop
		wp = (WindowPeek)LMGetWindowList();						// For each window in list,
		while (wp!=(WindowPeek)theWindow->win && wp!=NIL) {		//  up to window being closed,
			if (wp->visible)									//  if it's visible,
				DiffRgn(clipRgn,wp->strucRgn,clipRgn);			//  cut out its structure region
			wp = wp->nextWindow;
			}
		if (wp == NIL) goto cleanup;					// Should never happen

		// clipRgn now contains all pixels on desktop except those that belong
		// to windows in front of theWindow being closed.
		
		// Get maximum number of windows for which we might have to create entries
		
		numWindows = 1;
		wp = (WindowPeek)theWindow->win;
		while (wp!=(WindowPeek)theWidget->win && wp!=NIL) {
			if (wp->visible)
				numWindows++;
			wp = wp->nextWindow;
			}
		
		// Count how many intervening windows need evading in order to create
		// our zoom array.  There are always at least two entries in the array,
		// one for the start and one for the end ZoomRect.  If there are no
		// intervening windows, then this stuff will essentially do the same
		// thing as SingleSplineZoom.

		zoomArray = (ZoomRectHandle)NewHandle(numWindows * sizeof(ZoomRect));
		if (zoomArray == NIL) goto cleanup;
		
		// Set the first zoom to the starting ZoomRect for window
		
		(*zoomArray)[0] = *theWindow;
		(*zoomArray)[0].clip = clipRgn; clipRgn = NIL;		// Pass off clipRgn to first ZR
		
		// Set its knot to the center of its zoom rect bounds
		bounds = (*zoomArray)[0].zoom;
		(*zoomArray)[0].knot.x = (bounds.left + bounds.right) / 2;
		(*zoomArray)[0].knot.y = (bounds.top + bounds.bottom) / 2;
		
		// Traverse down the window list until we hit the widget's window
		numWindows = 1;
		wp = (WindowPeek)theWindow->win;
		if (wp) wp = wp->nextWindow;						// Skip first, we just did it above
		while (wp!=(WindowPeek)theWidget->win && wp!=NIL) {
			if (wp->visible) {
				// Check for whether the final destination is covered in any way by this window
				bounds = (*wp->strucRgn)->rgnBBox;
				if (SectRect(&theWidget->zoom,&bounds,&ans)) {
					// Intersection: need another knot in spline and intermediate ZR
					// Initialize the next intermediate ZoomRect's clipping region
					// as whatever we've got before, minus its window's structure.
					
					tmpRgn = NewRgn();
					(*zoomArray)[numWindows].clip = tmpRgn;
					if (tmpRgn)
						DiffRgn((*zoomArray)[numWindows-1].clip,wp->strucRgn,tmpRgn);
					 else
						goto cleanup;
					
					// Tell it which window it is evading, and keep same line width
					
					(*zoomArray)[numWindows].win = (WindowPtr)wp;
					(*zoomArray)[numWindows].width = theWindow->width;
					
					// As a test, we set the midway zooms to be on various sides of
					// the window frames, half size.
					
					w = (bounds.right - bounds.left);
					h = (bounds.bottom - bounds.top);
					
					// Basically, rotate our zoom around each successive window
					// This maximizes slinkyness and is very silly, but good for
					// demo purposes.  This is the spot in the routine where it
					// would be appropriate to analyze the window pattern as a
					// whole to try to optimize a simple path that evades the
					// group as a whole to get to the destination ZR.
					
					k = numWindows & 3;
					switch(k) {
						case 0: w = 0; h = -(h+16); break;	// 16 pixels above window
						case 1: w += 16; h = 0; break;		// 16 pixels to right of window
						case 2: w = 0; h += 16; break;		// 16 pixels below window
						case 3: w = -(w+16); h = 0; break;	// 16 pixels to left of window
						}
						
					OffsetRect(&bounds,w,h);		// Move it 16 pixels outside window
					// Make the thing smaller than entire window bounds
					InsetRect(&bounds,(bounds.right-bounds.left)/4,(bounds.bottom-bounds.top)/4);
					(*zoomArray)[numWindows].zoom = bounds;
					
					// Set knot to center of rect
					(*zoomArray)[numWindows].knot.x = (bounds.left + bounds.right) / 2;
					(*zoomArray)[numWindows].knot.y = (bounds.top + bounds.bottom) / 2;
					
					numWindows++;
					}
				}
			wp = wp->nextWindow;
			}
		
		// Set the last zoom to the ending ZoomRect for widget, and cut the array down
		// to its final size, since we allocated a maximum number of entries above.
		
		theWidget->knot.x = (theWidget->zoom.left + theWidget->zoom.right) / 2;
		theWidget->knot.y = (theWidget->zoom.top + theWidget->zoom.bottom) / 2;
		
		(*zoomArray)[numWindows] = *theWidget;
		tmpRgn = NewRgn();						// Use temporary so we don't have to
		(*zoomArray)[numWindows].clip = tmpRgn;	// lock zoomArray down during NewRgn.
		
		if (tmpRgn)
			CopyRgn((*zoomArray)[numWindows-1].clip,tmpRgn);
		 else
			goto cleanup;
		
		numWindows++;
		SetHandleSize((Handle)zoomArray,numWindows*sizeof(ZoomRect));
		
		// Finally, traverse the array, using line segments between knots to
		// choose spline control points that form a "smooth" path.
		// If the two control points on either side of a Bezier spline knot
		// are colinear, then the spline is continuous through the knot.
		// This means the zoom won't suddenly change direction as at passes
		// each individual ZoomRect entry in the zoomArray.  The hard part
		// is figuring out how to set the line that the two control points
		// have to be on so that the path isn't too crazy.
		// Note that if there are only 2 entries in the array, we don't
		// have any intermediate knots to worry about.
		// Naturally, this code would need to be sped up for older 68K macs
		// but it seems to work reasonably fast on the more recent machines
		// (trigonometry and linear algebra just to keep from confusing folks!)
		
		(*zoomArray)[0].c0 = (*zoomArray)[0].knot;
		
		for (k=1; k<(numWindows-1); k++) {
		
			double theta,dot,sn,cs; long a,b; Point2D pt;
			long dx0,dy0;			// Vector from last knot to this one
			long dx1,dy1;			// Vector from this knot to next one
			
			dx0 = (*zoomArray)[k].knot.x - (*zoomArray)[k-1].knot.x;
			dy0 = (*zoomArray)[k].knot.y - (*zoomArray)[k-1].knot.y;
			dx1 = (*zoomArray)[k+1].knot.x - (*zoomArray)[k].knot.x;
			dy1 = (*zoomArray)[k+1].knot.y - (*zoomArray)[k].knot.y;
			
			// Get the angle between the two vectors (this could be wrong: it's late)
			dot = dx0*dx1 + dy0+dy1;
			a = dx0*dx0 + dy0*dy0;
			b = dx1*dx1 + dy1*dy1;
			theta = acos( dot / (sqrt(a) * sqrt(b)) );
			
			// And the sin and cosine of half the angle
			theta = theta / 2.0;
			sn = sin(theta);
			cs = cos(theta);
			
			// Rotate our initial vector by half the angle, and shrink it to a third its size
			x = (cs*dx0 - sn*dy0);
			y = sn*dx0 + cs*dy0;
			x /= 3;							// Purely a heuristic: the larger this is,
			y /= 3;							// the more boisterous (wider) the turn will be
			
			// Set the two colinear control points for the next knot
			pt = (*zoomArray)[k].knot;
			(*zoomArray)[k-1].c1.x = pt.x - x;
			(*zoomArray)[k-1].c1.y = pt.y - y;
			(*zoomArray)[k].c0.x = pt.x + x;
			(*zoomArray)[k].c0.y = pt.y + y;
			}
		
		(*zoomArray)[numWindows-1].c1 = (*zoomArray)[numWindows-1].knot;
		
		okay = TRUE;		// Tell cleanup not to throw zoomArray away

cleanup:
		if (clipRgn) DisposeRgn(clipRgn);
		if (!okay) {
			if (zoomArray) DisposeHandle((Handle)zoomArray);
			zoomArray = NIL;
			}
		
		return(zoomArray);
	}

/*
 *	This routine creates a multiple set of Bezier curves that connect together
 *	to form a smooth path from start to end (if opening is TRUE) or from end to
 *	start (if opening is FALSE).  The path is then used to zoom a series of rects
 *	in a similar manner to the above SingleSplineZoom().  However, in this case
 *	the path can travel around on the desktop in order to avoid all intervening
 *	window frames between the window the widget is within and the window it has
 *	opened into.
 *
 *	It is possible for intervening windows to occur when opening is TRUE, even
 *	though the widget is nominally in the frontmost active window, because there
 *	may be floating palettes that zooming should avoid.
 *
 *	Note that before or after evading all intervening windows, the zoom still has
 *	to behave the same as SingleSplineZoom to take care of the case when the widget
 *	is not hidden in its own window.
 *
 *	The key to doing this is to define a set of intermediate rects alongside each
 *	intervening window, and zoom amongst them until it is time to dive into or out
 *	to the final window.  In addition, the clipping region has to be dynamically
 *	changed to add or delete various window structure regions so that the zoomrects
 *	get clipped as they move up or down the window list.
 *
 *	This routine disposes the zoomArray for the caller.
 */

void MultipleSplineZoom(ZoomRectHandle zoomArray, int opening)
	{
		short numWindows,i; ZoomRect start, end;
		
		if (zoomArray) {
			
			numWindows = GetHandleSize((Handle)zoomArray) / sizeof(ZoomRect);	// must be >= 2
			start = (*zoomArray)[0];
			end = (*zoomArray)[numWindows-1];
			
			// Break the two cases out while we're hacking.  If it looks like
			// they are basically the same, we'll put'em back together again.
			// Closing is more important for now.

			if (opening)
				SingleSplineZoom(&start,&end,opening);
			 else
				MultipleSplineZoomClose(zoomArray, &start, &end, numWindows);

			// Done, throw everything back into the primemordial soup
			
			for (i=0; i<numWindows; i++)
				if ((*zoomArray)[i].clip)
					DisposeRgn((*zoomArray)[i].clip);
					
			DisposeHandle((Handle)zoomArray);
			}
	}

/*
 *	theWindow->win has just been hidden, and this routine must zoom from
 *	its bounds down to the bounds of the widget within theWidget->win.
 */

void MultipleSplineZoomClose(ZoomRectHandle zoomArray, ZoomRect *theWindow, ZoomRect *theWidget, short numWindows)
	{
		Point startCenter,midCenter,endCenter,startSize,midSize,endSize,pt;
		Rect r1,r2,r3,r4,ans,midway,clip,content,widgetBounds;
		short zoomSteps,zoomSteps2,x,y,w,h,dx,dy,n,numPairs; long i,j,soon;
		Point2D p0,c1,c2,p3; int useDive;
		WindowPeek wp,*list; ZoomRect *start,*end;
		RgnHandle clip1=NIL,clip2=NIL,clip3=NIL,clip4=NIL,tmpClip;
		
		InstallDesktop();			// Draw outside of all windows; deskport becomes current
		
		// Fill the clipped rectangle queue with empty rectangles so nothing gets drawn
		// as the queue fills up.  The queue makes for a better zoom effect, since there
		// will be more graphic weight to the zooming, and the amount of time any
		// particular zoomrect is displayed is much longer.
		
  		SetRect(&r1,0,0,0,0);
		r2 = r1;
		r3 = r1;
		// Clipping regions 1 through 4 are allocated empty
		clip1 = NewRgn();			// Prepare to record clipping region for rectangle queue
		clip2 = NewRgn();
		clip3 = NewRgn();
		clip4 = NewRgn();
		if (clip4 == NIL) goto cleanup;
		
		// Now do a zoom between each pair of adjacent ZoomRects in table, of
		// which there are always at least one.
		
		HLock((Handle)zoomArray);
		
		start = (*zoomArray) + 0;
		end   = (*zoomArray) + 1;
		numPairs = numWindows - 1;
		
		// Each iteration slides start/end up by one in the array
		
		for (n=0; n<numPairs; n++,start++,end++) {
		
	  		// Get centers of the two limiting rectangles in this zoom segment
	  		
			startCenter.h = (start->zoom.left + start->zoom.right) / 2;
			startCenter.v = (start->zoom.top + start->zoom.bottom) / 2;
			endCenter.h = (end->zoom.left + end->zoom.right) / 2;
			endCenter.v = (end->zoom.top + end->zoom.bottom) / 2;
	
			// And the starting and ending half sizes
			
			startSize.h = (start->zoom.right - start->zoom.left) / 2;
			startSize.v = (start->zoom.bottom - start->zoom.top) / 2;
			endSize.h = (end->zoom.right - end->zoom.left) / 2;
			endSize.v = (end->zoom.bottom - end->zoom.top) / 2;
			
			// Get knot points on spline in global coordinates
			p0.x = startCenter.h; p0.y = startCenter.v;
			p3.x = endCenter.h; p3.y = endCenter.v;
			
			// If this is final widget window, then do like SingleSplineZoom above
			
			if (n == (numPairs-1)) {
			
				// If either start or end rectangle is hidden in any way, then we have
				// to use a curved path to "get inside/outside" the window it's hidden in.
				// Have to convert to local coords because portRect is in local coords.
				// Since a line is a special case of a Bezier curve when the knots and
				// control points are co-linear, we can use the spline code to do a
				// linear zoom when both start and end are visible.
			
				useDive = TRUE;
				GlobalToLocalZR(theWidget);
				if (SectRect(&theWidget->win->portRect,&theWidget->zoom,&ans))
					if (EqualRect(&ans,&theWidget->zoom)) {
						// Widget at start or end is completely visible
						// so we can use control points that force the
						// path to be a line and avoid clipping stuff
						widgetBounds = theWidget->zoom;			// Save a copy in local coords
						useDive = FALSE;
						}
				LocalToGlobalZR(theWidget);
				
				if (useDive) {
				
					PushPort(theWidget->win);		// Prepare for coordinate conversions
					
					// Get content area minus the scroll bar and grow icon areas
					// This assumes window has both right and bottom scroll bars
					content = theWidget->win->portRect;
					content.right -= SCROLLBARWIDTH;
					content.bottom -= SCROLLBARWIDTH;
					
					// Find a rectangle within the window to serve as an intermediate
					// destination for the zoom that is completely contained in the
					// visible area of the destination zoom window
					midway.left = midway.top = 0;
					midway.right = (content.right + (theWidget->zoom.right-theWidget->zoom.left)) / 2;
					midway.bottom = (content.bottom + (theWidget->zoom.bottom-theWidget->zoom.top)) / 2;
					CenterRect(&midway,&content,&midway);
					
					// If midway is larger than window, just use inset window content bounds
					if (midway.left<0 || midway.top<0 ||
							midway.right>content.right || midway.bottom>content.bottom) {
						midway = content;
						InsetRect(&midway,8,8);
						}
					
					// Convert to global coords
					pt.h = midway.left; pt.v = midway.top;
					LocalToGlobal(&pt);
					midway.left = pt.h; midway.top = pt.v;
					pt.h = midway.right; pt.v = midway.bottom;
					LocalToGlobal(&pt);
					midway.right = pt.h; midway.bottom = pt.v;
					midSize.h = (midway.right - midway.left) / 2;
					midSize.v = (midway.bottom - midway.top) / 2;
					midCenter.h = (midway.right + midway.left) / 2;
					midCenter.v = (midway.bottom + midway.top) / 2;
					
					// Get window content as clipping rectangle in global coords
					pt.h = pt.v = 0;
					LocalToGlobal(&pt);
					clip.left = pt.h; clip.top = pt.v;
					pt.h = content.right;
					pt.v = content.bottom;
					LocalToGlobal(&pt);
					clip.right = pt.h; clip.bottom = pt.v;
					
					// But have to take intersection of it with final window clipping region
					
					tmpClip = NewRgn();
					if (tmpClip) {
						RectRgn(tmpClip,&clip);
						SectRgn(end->clip,tmpClip,end->clip);
						DisposeRgn(tmpClip);
						}
					
					// Set control (tension) points to other side of window in global coordinates
					pt.h = content.right / 2;
					pt.v = content.bottom / 2;

					if (end->zoom.bottom > midway.bottom) pt.v = CONTROLINSET;
					if (end->zoom.top < midway.top) pt.v = content.bottom - CONTROLINSET;
					if (end->zoom.right > midway.right) pt.h = CONTROLINSET;
					if (end->zoom.left < midway.left) pt.h = content.right - CONTROLINSET;

					LocalToGlobal(&pt);
					c1.x = c2.x = pt.h;
					c1.y = c2.y = pt.v;
		
					PopPort();
					}
				 else {
					// Standard zoom: Set control points to 1/2 way between start and end points
					c1.x = c2.x = (p0.x + p3.x) / 2;
					c1.y = c2.y = (p0.y + p3.y) / 2;
					}
				}
			 else {
				// Not final widget window: still evading window frames
				useDive = FALSE;
				// Use precomputed control points for smooth turns through the knot
				p0 = start->knot;
				p3 = end->knot;
				c1 = start->c0;
				c2 = start->c1;
				}
			
			// Precompute the spline's path, including both endpoints
			
			ComputeBezierSpline(&p0,&c1,&c2,&p3,path,MAXPATH);
			
			// You could interpolate among bounding pen sizes if you wanted to give
			// more of a depth effect.
			PenSize(start->width,start->width);
			
			zoomSteps = MAXPATH;
			zoomSteps2 = zoomSteps/2;				// Halfway point
			
			for (i=0; i<=zoomSteps; i++) {			// loops zoomSteps+1 times
			
				x = path[i].h;
				y = path[i].v;
	
				if (useDive) {
					if (i <= zoomSteps2) {
						// First half of zoom: interpolate from start to midway on desktop
						w = ((zoomSteps2-i)*startSize.h + i*midSize.h) / zoomSteps2;
						h = ((zoomSteps2-i)*startSize.v + i*midSize.v) / zoomSteps2;
						SetClip(start->clip);
						}
					 else {
					 	// Second half of zoom: interpolate from midway to end within window
					 	j = i - zoomSteps2;
						w = ((zoomSteps2-j)*midSize.h + j*endSize.h) / zoomSteps2;
						h = ((zoomSteps2-j)*midSize.v + j*endSize.v) / zoomSteps2;
						SetClip(end->clip);		// Install window content clipping
						}
					}
				 else {
					// Get i'th intermediate size (interpolated linearly) for whole zoom
					w = ((zoomSteps-i)*startSize.h + i*endSize.h) / zoomSteps;
					h = ((zoomSteps-i)*startSize.v + i*endSize.v) / zoomSteps;
					// Set clipping to exclude all higher windows
					SetClip(start->clip);
					}
				
				// Build the next interpolated rectangle in queue and draw it
				GetClip(clip4);
				SetRect(&r4,x-w,y-h,x+w,y+h);
				FrameRect(&r4);
				
				// Erase (assuming xor mode) the i-3'rd previously drawn rectangle
				SetClip(clip1);
				FrameRect(&r1);
				
				// Shift rectangle queue up by 1, leaving r4 ready to be redefined
				r1 = r2; r2 = r3; r3 = r4;
				tmpClip = clip1; clip1 = clip2; clip2 = clip3; clip3 = clip4; clip4 = tmpClip;
				
				// Use governor so processor speed doesn't affect zoom
				Wait(1);
				}
			}
		
		HUnlock((Handle)zoomArray);
		
		// Erase last three zoomrects to empty the queue of drawn zoomrects
		
		SetClip(clip1); FrameRect(&r1);
		SetClip(clip2); FrameRect(&r2);
		SetClip(clip3); FrameRect(&r3);
		
cleanup:
		if (clip4) DisposeRgn(clip4);
		if (clip3) DisposeRgn(clip3);
		if (clip2) DisposeRgn(clip2);
		if (clip1) DisposeRgn(clip1);
		
		ResetDesktop();			// Restore normal drawing environment
	}

/*
 *	Compute, in global coordinates, the path of a Bezier spline in 2D coordinates
 *	whose starting and ending points are p0 and p3, and whose control points are
 *	c1 and c2.  The path should be stored in the array "path", which is expected
 *	to be able to hold (numPoints+1) elements, from 0 to numPoints, inclusive.
 *	numPoints should be a power of 2 between 2 and 32, inclusive.  This routine
 *	can be easily generalized to 3D (or higher) coordinates, and is optimized
 *	for fast computation in (long) integers.  Since there are no divides, the
 *	routine does the right thing regardless of whether p0, c1, c2, or p3 are
 *	all different or coincident or whatever.
 */

void ComputeBezierSpline(Point2D *p0, Point2D *c1, Point2D *c2, Point2D *p3,
													Point *path, short numPoints)
	{		
		long i,ax,ay,az,bx,by,bz,cx,cy,cz,curx,cury,curz;
		short s,s1,s2,s3;

		curx = p0->x; cury = p0->y;

		//	Compute the integer Bezier spline coefficients, a, b, and c

		cx = (c1->x - curx); cx += cx << 1;			// times 3
		cy = (c1->y - cury); cy += cy << 1;
		
		bx = (c2->x - c1->x); bx += (bx << 1) - cx;	// times 3 - c
		by = (c2->y - c1->y); by += (by << 1) - cy;
		
		ax = (p3->x - curx) - cx - bx;
		ay = (p3->y - cury) - cy - by;

		s = numPoints;							// Number of lines in spline, and bits
		if		 (numPoints == 32) s1 = 5;
		 else if (numPoints == 16) s1 = 4;
		 else if (numPoints ==  8) s1 = 3;
		 else if (numPoints ==  4) s1 = 2;
		 else					 { s1 = 1; s = 0; }
		
		s2 = s1+s1; s3 = s2+s1;		// s3 is 15 bits worth of scaling
		
		bx   <<= s1;   by <<= s1;	// Scale operands up for later, according
		cx   <<= s2;   cy <<= s2;	// to the degree in i in loop below
		curx <<= s3; cury <<= s3;

		for (i=0; i<=s; i++) {
			path[i].h = (i * (i * (i * ax + bx) + cx) + curx) >> s3;
			path[i].v = (i * (i * (i * ay + by) + cy) + cury) >> s3;
			}
  }

